@@ -2,7 +2,7 @@
 Module for parallel flow training and execution.
 
 Not that this module depends on bihinet, since it uses a BiFlowNode to
-encapsulate the BiFlow in the tasks. 
+encapsulate the BiFlow in the tasks.
 """
 
 import itertools
@@ -12,181 +12,97 @@ n = mdp.numx
 
 import mdp.parallel as parallel
 
-from ..biflow import (BiFlow, BiFlowException, MessageResultContainer,
-                      BiCheckpointFlow)
-
-from parallelbihinet import BiFlowNode
+from bimdp import (
+    BiFlow, BiFlowException, MessageResultContainer, BiCheckpointFlow,
+    EXIT_TARGET
+)
+from bimdp.hinet import BiFlowNode
 
 
 ### Train Task Classes ###
 
-class BiFlowTrainTaskException(Exception):
-    """Exception for problems with the BiFlowTrainTask execution."""
-    pass
-    
-    
-class BiFlowTrainCallable(parallel.FlowTaskCallable):
+class BiFlowTrainCallable(parallel.FlowTrainCallable):
     """Task implementing a single training phase in a flow for a data block."""
 
-    def __init__(self, biflownode, purge_nodes=True):
-        """Store everything for the training.
-        
-        biflownode -- BiFlowNode encapsulating the forked BiFlow.
-        purge_nodes -- If True nodes not needed for the join will be replaced
-            with dummy nodes to reduce the footprint.
-        """
-        self._biflownode = biflownode
-        self._purge_nodes = purge_nodes
-        super(BiFlowTrainCallable, self).__init__()
-        
     def __call__(self, data):
         """Do the training and return the purged BiFlowNode.
-        
+
         data -- tuple containing x and msg
         """
         x, msg = data
         while True:
-            result = self._biflownode.train(x, msg)
+            result = self._flownode.train(x, msg)
             if (result is None) or isinstance(result, dict):
                 break
-            elif len(result) == 4:
-                # discard global message and reenter
-                x, msg = result[:2]
+            elif (isinstance(result, tuple) and
+                (result[2] in [1, -1, EXIT_TARGET])):
+                break
             else:
                 err = ("Target node not found in flow during " +
                        "training, last result: " + str(result))
                 raise BiFlowException(err)
-        self._biflownode.bi_reset()
+        self._flownode.bi_reset()
         if self._purge_nodes:
-            self._biflownode.purge_nodes()
-        return self._biflownode
-    
+            parallel._purge_flownode(self._flownode)
+        return self._flownode
+
     def fork(self):
-        return self.__class__(self._biflownode.fork(),
+        return self.__class__(self._flownode.fork(),
                               purge_nodes=self._purge_nodes)
-    
-    
-class BiFlowTrainResultContainer(parallel.ResultContainer):
-    """Container for a BiFlowNode.
+
 
-    This class is required for parallel BiFlow training since there may be
-    nodes for which is_bi_training is True during the normal training phase.
-    """
-    
-    def __init__(self):
-        self._biflownode = None
-        
-    def add_result(self, result, task_index):
-        if not self._biflownode:
-            self._biflownode = result
-        else:
-            self._biflownode.join(result)
-            
-    def get_results(self):
-        biflownode = self._biflownode
-        self._biflownode = None
-        return [biflownode,]
-    
-    
 ### Execute Task Classes ###
 
-class BiFlowExecuteTaskException(Exception):
-    """Exception for problems with the BiFlowExecuteTask execution."""
-    pass
-    
-    
-class BiFlowExecuteCallable(parallel.FlowTaskCallable):
+class BiFlowExecuteCallable(parallel.FlowExecuteCallable):
     """Task implementing data execution for a BiFlowNode."""
-
-    def __init__(self, biflownode, purge_nodes=True):
+    
+    def __init__(self, flownode, purge_nodes=True):
         """Store everything for the execution.
-        
-        biflownode -- BiFlowNode for the execution
+
+        flownode -- FlowNode for the execution
         purge_nodes -- If True nodes not needed for the join will be replaced
             with dummy nodes to reduce the footprint.
         """
-        self._biflownode = biflownode
-        self._purge_nodes = purge_nodes
-        super(BiFlowExecuteCallable, self).__init__()
-        
+        super(BiFlowExecuteCallable, self).__init__(flownode,
+                                                    purge_nodes=purge_nodes)
+
     def __call__(self, data):
         """Return the execution result and the BiFlowNode as a tuple.
-        
-        If is_bi_training() is False for the BiFlowNode then None is returned
-        instead of the BiFlowNode. If is_bi_training() is True then the 
-        BiFlowNode is purged.
+
+        If use_fork_execute is True for the flownode then it is returned
+        in the result tuple.
         """
         x, msg, target = data
         # by using _flow we do not have to reenter (like for train)
-        result = self._biflownode._flow.execute(x, msg, target)
-        self._biflownode.bi_reset()
-        if self._biflownode.is_bi_training():
+        result = self._flownode._flow.execute(x, msg, target)
+        self._flownode.bi_reset()
+        if self._flownode.use_execute_fork():
             if self._purge_nodes:
-                self._biflownode.purge_nodes()
-            return (result, self._biflownode)
+                parallel._purge_flownode(self._flownode)
+            return (result, self._flownode)
         else:
             return (result, None)
-        
+
     def fork(self):
-        return self.__class__(self._biflownode.fork(),
+        return self.__class__(self._flownode.fork(),
                               purge_nodes=self._purge_nodes)
 
 
-class OrderedBiExecuteResultContainer(parallel.OrderedResultContainer):
-    """Default result container with automatic restoring of the result order.
-    
-    This result container should be used together with BiFlowExecuteCallable.
-    Both the execute result (x and possibly msg) and the forked BiFlowNode
-    are stored.
-    """
-    
-    def __init__(self):
-        """Initialize attributes."""
-        super(OrderedBiExecuteResultContainer, self).__init__()
-        self._biflownode = None
-    
-    def add_result(self, result, task_index):
-        """Remove the forked BiFlowNode from the result and join it."""
-        excecute_result, forked_biflownode = result
-        super(OrderedBiExecuteResultContainer, self).add_result(excecute_result, 
-                                                                task_index)
-        # join biflownode
-        if forked_biflownode is not None:
-            if self._biflownode is None:
-                self._biflownode = forked_biflownode
-            else:
-                self._biflownode.join(forked_biflownode)
-    
-    def get_results(self):
-        """Return the ordered results.
-        
-        The joined BiFlowNode is returned in the first result list entry,
-        for the following result entries BiFlowNode is set to None.
-        This reduces memory consumption while staying transparent for the
-        ParallelBiFlow.
-        """
-        excecute_results = super(OrderedBiExecuteResultContainer, 
-                                                            self).get_results()
-        biflownode_results = ([self._biflownode,] 
-                              + ([None] * (len(excecute_results)-1)))
-        return zip(excecute_results, biflownode_results)
-    
-    
 ### ParallelBiFlow Class ###
 
 class ParallelBiFlowException(parallel.ParallelFlowException):
     """Standard exception for problems with ParallelBiFlow."""
-    
+
 
 class ParallelBiFlow(BiFlow, parallel.ParallelFlow):
     """A parallel  provides the tasks for parallel training.
-    
+
     Note that even though a node input x or output y can be None, the data
-    iterables cannot be None themselves, since they define the iterator length 
+    iterables cannot be None themselves, since they define the iterator length
     for the message iterator as well. They can, however, return None for each
     iteration step.
     """
-    
+
     def __init__(self, flow, verbose=False, **kwargs):
         """Initialize the internal variables."""
         self._train_msg_iterables = None
@@ -195,19 +111,19 @@ class ParallelBiFlow(BiFlow, parallel.ParallelFlow):
         self._exec_msg_iterator = None
         self._exec_target_iterator = None
         super(ParallelBiFlow, self).__init__(flow, verbose=verbose, **kwargs)
-    
+
     @mdp.with_extension("parallel")
-    def train(self, data_iterables, msg_iterables=None, 
+    def train(self, data_iterables, msg_iterables=None,
               stop_messages=None,
               scheduler=None,
               train_callable_class=None,
               overwrite_result_container=True,
               **kwargs):
         """Parallel version of the standard train method.
-        
+
         If a scheduler is provided the training will be done in parallel on the
         scheduler.
-        
+
         data_iterables -- A list of iterables, one for each node in the flow.
             The iterators returned by the iterables must
             return data arrays that are then used for the node training.
@@ -216,8 +132,8 @@ class ParallelBiFlow(BiFlow, parallel.ParallelFlow):
             then other data types can be used as well.
         msg_iterables - A list of iterables for the messages.
         stop_messages -- Sequence of messages for stop_training.
-        scheduler -- Value can be either None for normal training (default 
-            value) or a Scheduler instance for parallel training with the 
+        scheduler -- Value can be either None for normal training (default
+            value) or a Scheduler instance for parallel training with the
             scheduler.
             If the scheduler value is an iterable or iterator then it is
             assumed that it contains a scheduler for each training phase.
@@ -225,10 +141,10 @@ class ParallelBiFlow(BiFlow, parallel.ParallelFlow):
             you can e.g. use a generator to create the schedulers just in time.
             For nodes which are not trained the scheduler can be None.
         train_callable_class -- Class used to create training callables for the
-            scheduler. By specifying your own class you can implement data 
-            transformations before the data is actually fed into the flow 
-            (e.g. from 8 bit image to 64 bit double precision). 
-            Note that the train_callable_class is only used if a scheduler was 
+            scheduler. By specifying your own class you can implement data
+            transformations before the data is actually fed into the flow
+            (e.g. from 8 bit image to 64 bit double precision).
+            Note that the train_callable_class is only used if a scheduler was
             provided. If a scheduler is provided the default class used is
             NodeResultContainer.
         overwrite_result_container -- If set to True (default value) then
@@ -253,7 +169,7 @@ class ParallelBiFlow(BiFlow, parallel.ParallelFlow):
             # do parallel training
             try:
                 self.setup_parallel_training(
-                                        data_iterables=data_iterables, 
+                                        data_iterables=data_iterables,
                                         msg_iterables=msg_iterables,
                                         stop_messages=stop_messages,
                                         train_callable_class=train_callable_class,
@@ -284,8 +200,9 @@ class ParallelBiFlow(BiFlow, parallel.ParallelFlow):
                 if ((scheduler is not None) and
                     overwrite_result_container and
                     (not isinstance(scheduler.result_container,
-                                    BiFlowTrainResultContainer))):
-                    scheduler.result_container = BiFlowTrainResultContainer()
+                                    parallel.TrainResultContainer))):
+                    scheduler.result_container = \
+                        parallel.TrainResultContainer()
                 while self.is_parallel_training:
                     while self.task_available:
                         task = self.get_task()
@@ -310,27 +227,27 @@ class ParallelBiFlow(BiFlow, parallel.ParallelFlow):
                         if ((scheduler is not None) and
                             overwrite_result_container and
                             (not isinstance(scheduler.result_container,
-                                            BiFlowTrainResultContainer))):
+                                            parallel.TrainResultContainer))):
                             scheduler.result_container = \
-                                            BiFlowTrainResultContainer()
+                                            parallel.TrainResultContainer()
             finally:
                 # reset remaining iterator references, which cannot be pickled
-                self._train_data_iterator = None    
+                self._train_data_iterator = None
                 self._train_msg_iterator = None
                 if (schedulers is not None) and (scheduler is not None):
                     scheduler.shutdown()
-    
-    def setup_parallel_training(self, data_iterables, msg_iterables=None, 
+
+    def setup_parallel_training(self, data_iterables, msg_iterables=None,
                                 stop_messages=None,
                                 train_callable_class=BiFlowTrainCallable):
         """Prepare the flow for handing out tasks to do the training.
-        
+
         After calling setup_parallel_training one has to pick up the
         tasks with get_task, run them and finally return the results via
-        use_results. tasks are available as long as task_available is 
-        True. Training may require multiple phases, which are each closed by 
+        use_results. tasks are available as long as task_available is
+        True. Training may require multiple phases, which are each closed by
         calling use_results.
-        
+
         data_iterables -- A list of iterables, one for each node in the flow.
             The iterators returned by the iterables must
             return data arrays that are then used for the node training.
@@ -338,13 +255,13 @@ class ParallelBiFlow(BiFlow, parallel.ParallelFlow):
             If a custom train_callable_class is used to preprocess the data
             then other data types can be used as well.
         msg_iterables - A list of iterables for the messages. Can also be
-            a single message if data_iterables is a single array. 
+            a single message if data_iterables is a single array.
         stop_messages -- Sequence of messages for stop_training.
         train_callable_class -- Class used to create training callables for the
-            scheduler. By specifying your own class you can implement data 
-            transformations before the data is actually fed into the flow 
-            (e.g. from 8 bit image to 64 bit double precision). 
-            Note that the train_callable_class is only used if a scheduler was 
+            scheduler. By specifying your own class you can implement data
+            transformations before the data is actually fed into the flow
+            (e.g. from 8 bit image to 64 bit double precision).
+            Note that the train_callable_class is only used if a scheduler was
             provided. If a scheduler is provided the default class used is
             NodeResultContainer.
         """
@@ -354,7 +271,7 @@ class ParallelBiFlow(BiFlow, parallel.ParallelFlow):
             raise ParallelBiFlowException(err)
         self._train_callable_class = train_callable_class
         data_iterables, msg_iterables = self._sanitize_training_iterables(
-                                            data_iterables=data_iterables, 
+                                            data_iterables=data_iterables,
                                             msg_iterables=msg_iterables)
         self._train_data_iterables = data_iterables
         self._train_msg_iterables = msg_iterables
@@ -364,12 +281,12 @@ class ParallelBiFlow(BiFlow, parallel.ParallelFlow):
         self._flownode = BiFlowNode(BiFlow(self.flow))
         self._i_train_node = 0
         self._next_train_phase()
-            
+
     def _next_train_phase(self):
         """Find the next phase or node for parallel training.
-        
+
         When it is found the corresponding internal variables are set.
-        Nodes which are not derived from ParallelNode are trained locally. 
+        Nodes which are not derived from ParallelNode are trained locally.
         If a fork() fails due to a TrainingPhaseNotParallelException
         in a certain train phase, then the training is done locally as well
         (but fork() is tested again for the next phase).
@@ -382,14 +299,14 @@ class ParallelBiFlow(BiFlow, parallel.ParallelFlow):
                 continue
             iterable = self._train_data_iterables[self._i_train_node]
             msg_iterable = self._train_msg_iterables[self._i_train_node]
-            iterable, msg_iterable, _ = self._sanitize_iterables(iterable, 
+            iterable, msg_iterable, _ = self._sanitize_iterables(iterable,
                                                                  msg_iterable)
             try:
                 self._flownode.fork()
                 # fork successful, prepare parallel training
                 if self.verbose:
                     print ("start parallel training phase of " +
-                           "node no. %d in parallel flow" % 
+                           "node no. %d in parallel flow" %
                            (self._i_train_node+1))
                 self._train_data_iterator = iter(iterable)
                 self._train_msg_iterator = iter(msg_iterable)
@@ -411,63 +328,65 @@ class ParallelBiFlow(BiFlow, parallel.ParallelFlow):
                 if task_data_chunk is None:
                     err = "Training data iterator is empty."
                     raise ParallelBiFlowException(err)
-                # first task contains the new callable
+                # Only first task contains the new callable (enable caching).
+                # A fork is not required here, since the callable is always
+                # forked in the scheduler.
                 self._next_task = (task_data_chunk,
-                            self._train_callable_class(self._flownode.fork(),
+                            self._train_callable_class(self._flownode,
                                                        purge_nodes=True))
                 break
-            except parallel.TrainingPhaseNotParallelException, e:
+            except parallel.NotForkableParallelException, exception:
                 if self.verbose:
                     print ("could not fork node no. %d: %s" %
-                           (self._i_train_node + 1, str(e)))
-                    print ("start nonparallel training phase of " + 
-                           "node no. %d in parallel flow" % 
+                           (self._i_train_node + 1, str(exception)))
+                    print ("start nonparallel training phase of " +
+                           "node no. %d in parallel flow" %
                            (self._i_train_node+1))
                 self._local_train_phase(iterable, msg_iterable)
                 if self.verbose:
-                    print ("finished nonparallel training phase of " + 
-                           "node no. %d in parallel flow" % 
+                    print ("finished nonparallel training phase of " +
+                           "node no. %d in parallel flow" %
                            (self._i_train_node+1))
                 if not self.flow[self._i_train_node].is_training():
                     self._i_train_node += 1
         else:
             # training is finished
             self._i_train_node = None
-            
+
     def _local_train_phase(self, iterable, msg_iterable):
         """Perform a single training phase locally.
-        
+
         The internal _train_callable_class is used for the training.
         """
         task_callable = self._train_callable_class(self._flownode,
                                                    purge_nodes=False)
-        i_task = 0 
+        i_task = 0
         for (x, msg) in itertools.izip(iterable, msg_iterable):
             i_task += 1
             # Note: if x contains additional args assume that the
-            # callable can handle this  
+            # callable can handle this
             task_callable((x, msg))
             if self.verbose:
                 print ("    finished nonparallel task no. %d" % i_task)
-        if self.verbose:
-            print ("finished nonparallel training phase of " + 
-                   "node no. %d in parallel flow" % 
-                   (self._i_train_node+1))
         # perform stop_training with result check
         self._stop_training_hook()
         result = self._flownode.stop_training(
                                 self._stop_messages[self._i_train_node])
         self._post_stop_training_hook()
         if (result is not None) and (not isinstance(result, dict)):
-            err = ("Target node not found in flow during " +
-                   "stop_message propagation, last result: " + 
-                   str(result))
-            raise BiFlowException(err)
+            if (isinstance(result, tuple) and
+                (result[2] in [1, -1, EXIT_TARGET])):
+                pass
+            else:
+                err = ("Target node not found in flow during " +
+                       "stop_training phase, last result: " +
+                       str(result))
+                raise BiFlowException(err)
         self._bi_reset()
-            
+
     def _create_train_task(self):
         """Create and return a single training task without callable.
-        
+
         Returns None if data iterator end is reached.
         Raises NoTaskException if any other problem arises.
         """
@@ -477,33 +396,33 @@ class ParallelBiFlow(BiFlow, parallel.ParallelFlow):
             return ((x, msg), None)
         except StopIteration:
             return None
-    
-    @mdp.with_extension("parallel")  # needed for fork in local scheduler 
+
+    @mdp.with_extension("parallel")  # needed for fork in local scheduler
     def execute(self, iterable=None, msg_iterable=None, target_iterable=None,
-                scheduler=None, 
+                scheduler=None,
                 execute_callable_class=None,
                 overwrite_result_container=True):
-        """Train all trainable nodes in the flow.
-        
+        """Execute the flow and return (y, msg).
+
         If a scheduler is provided the execution will be done in parallel on
         the scheduler.
-        
+
         iterable -- Single array or iterable.
         msg_iterable -- Single message or iterable.
         target_iterable -- Single target or iterable.
-        scheduler -- Value can be either None for normal execution (default 
-            value) or a Scheduler instance for parallel execution with the 
+        scheduler -- Value can be either None for normal execution (default
+            value) or a Scheduler instance for parallel execution with the
             scheduler.
-        execute_callable_class -- Class used to create execution callables for 
-            the scheduler. By specifying your own class you can implement data 
-            transformations before the data is actually fed into the flow 
-            (e.g. from 8 bit image to 64 bit double precision). 
-            Note that the execute_callable_class is only used if a scheduler was 
+        execute_callable_class -- Class used to create execution callables for
+            the scheduler. By specifying your own class you can implement data
+            transformations before the data is actually fed into the flow
+            (e.g. from 8 bit image to 64 bit double precision).
+            Note that the execute_callable_class is only used if a scheduler was
             provided. If a scheduler is provided the default class used is
             NodeResultContainer.
         overwrite_result_container -- If set to True (default value) then
             the result container in the scheduler will be overwritten with an
-            instance of OrderedResultContainer, if it is not already an 
+            instance of OrderedResultContainer, if it is not already an
             instance of OrderedResultContainer.
         """
         if self.is_parallel_training:
@@ -521,12 +440,13 @@ class ParallelBiFlow(BiFlow, parallel.ParallelFlow):
         # check that the scheduler is compatible
         if overwrite_result_container:
             if not isinstance(scheduler.result_container,
-                              OrderedBiExecuteResultContainer):
-                scheduler.result_container = OrderedBiExecuteResultContainer()
+                              parallel.ExecuteResultContainer):
+                scheduler.result_container = parallel.ExecuteResultContainer()
         # do parallel execution
+        self._flownode = BiFlowNode(BiFlow(self.flow))
         try:
             self.setup_parallel_execution(
-                                iterable=iterable, 
+                                iterable=iterable,
                                 msg_iterable=msg_iterable,
                                 target_iterable=target_iterable,
                                 execute_callable_class=execute_callable_class)
@@ -540,38 +460,32 @@ class ParallelBiFlow(BiFlow, parallel.ParallelFlow):
             self._exec_msg_iterator = None
             self._exec_target_iterator = None
         return result
-            
+
     def setup_parallel_execution(self, iterable, msg_iterable=None,
                                  target_iterable=None,
                                  execute_callable_class=BiFlowExecuteCallable):
-        """Parallel version of the standard execute method.
-        
+        """Prepare the flow for handing out tasks to do the execution.
+
         Instead of automatically executing the _flow with the iterable, it only
         prepares the tasks for the scheduler.
-        
+
         iterable -- Single array or iterable.
         msg_iterable -- Single message or iterable.
         target_iterable -- Single target or iterable.
-        execute_callable_class -- Class used to create execution callables for 
-            the scheduler. By specifying your own class you can implement data 
-            transformations before the data is actually fed into the flow 
-            (e.g. from 8 bit image to 64 bit double precision). 
-            Note that the execute_callable_class is only used if a scheduler was 
-            provided. If a scheduler is provided the default class used is
+        execute_callable_class -- Class used to create execution callables for
+            the scheduler. By specifying your own class you can implement data
+            transformations before the data is actually fed into the flow
+            (e.g. from 8 bit image to 64 bit double precision).
+            Note that the execute_callable_class is only used if a scheduler
+            was provided. If a scheduler is provided the default class used is
             NodeResultContainer.
         """
         self._bi_reset()  # normally not required, just for safety
         if self.is_parallel_training:
             raise ParallelBiFlowException("Parallel training is underway.")
-        self._flownode = BiFlowNode(BiFlow(self.flow))
-        if self._flownode.is_bi_training():
-            # this will raise an exception if fork is not supported
-            task_flownode = self._flownode.fork()
-        else:
-            task_flownode = self._flownode.copy()
         self._execute_callable_class = execute_callable_class
         iterable, msg_iterable, target_iterable = self._sanitize_iterables(
-                                                           iterable, 
+                                                           iterable,
                                                            msg_iterable,
                                                            target_iterable)
         self._exec_data_iterator = iter(iterable)
@@ -585,14 +499,16 @@ class ParallelBiFlow(BiFlow, parallel.ParallelFlow):
         if task_data_chunk is None:
             err = "Execution data iterable is empty."
             raise ParallelBiFlowException(err)
-        # first task contains the new callable
+        # Only first task contains the new callable (enable caching).
+        # A fork is not required here, since the callable is always
+        # forked in the scheduler.
         self._next_task = (task_data_chunk,
-                           self._execute_callable_class(task_flownode,
+                           self._execute_callable_class(self._flownode,
                                                         purge_nodes=True))
-    
+
     def _create_execute_task(self):
         """Create and return a single execution task.
-        
+
         Returns None if data iterator end is reached.
         Raises NoTaskException if no task is available.
         """
@@ -603,17 +519,17 @@ class ParallelBiFlow(BiFlow, parallel.ParallelFlow):
             return ((x, msg, target), None)
         except StopIteration:
             return None
-    
+
     def use_results(self, results):
         """Use the result from the scheduler.
-        
+
         During parallel training this will start the next training phase.
-        For parallel execution this will return the result, like a normal 
+        For parallel execution this will return the result, like a normal
         execute would. In addition it will join any forked nodes.
-        
+
         results -- Iterable containing the results, normally the return value
-            of scheduler.ResultContainer.get_results().  
-            The individual results can be the return values of the tasks. 
+            of scheduler.ResultContainer.get_results().
+            The individual results can be the return values of the tasks.
         """
         if self.is_parallel_training:
             for result in results:
@@ -624,20 +540,16 @@ class ParallelBiFlow(BiFlow, parallel.ParallelFlow):
                                     self._stop_messages[self._i_train_node])
             self._post_stop_training_hook()
             if (result is not None):
-                target = result[1]
-                # values of +1 and -1 are tolerated
-                if isinstance(target, int) and (target == 1 or target == -1):
-                    pass
-                else:
+                target = result[2]
+                # values of +1, -1 and EXIT_TARGET are tolerated
+                if target not in [1, -1, EXIT_TARGET]:
                     err = ("Target node not found in flow during " +
-                           "stop_message propagation, last result: " + 
+                           "stop_training phase, last result: " +
                            str(result))
                     raise BiFlowException(err)
             self._flownode.bi_reset()
-            # update the node list of this flow
-            self.flow = self._flownode._flow.flow
             if self.verbose:
-                print ("finished parallel training phase of node no. " + 
+                print ("finished parallel training phase of node no. " +
                        "%d in parallel flow" % (self._i_train_node+1))
             if not self.flow[self._i_train_node].is_training():
                 self._i_train_node += 1
@@ -650,7 +562,6 @@ class ParallelBiFlow(BiFlow, parallel.ParallelFlow):
             msg_results = MessageResultContainer()
             # use internal flownode to join all biflownodes
             self._flownode = BiFlowNode(BiFlow(self.flow))
-            did_join_flownodes = False  # flag to show if a join took place
             for result_tuple in results:
                 result, forked_biflownode = result_tuple
                 # consolidate results
@@ -670,12 +581,6 @@ class ParallelBiFlow(BiFlow, parallel.ParallelFlow):
                 # join biflownode
                 if forked_biflownode is not None:
                     self._flownode.join(forked_biflownode)
-                    did_join_flownodes = True
-            # update the node list of this flow
-            self.flow = self._flownode._flow.flow
-            if self.verbose and did_join_flownodes:
-                print ("joined nodes with forked nodes from " + 
-                       "parallel execution")
             # return results
             if y_results is not None:
                 y_results = n.concatenate(y_results)
@@ -684,22 +589,22 @@ class ParallelBiFlow(BiFlow, parallel.ParallelFlow):
             err = "It seems that there are no results to retrieve."
             raise BiFlowException(err)
 
-    
+
 class ParallelCheckpointBiFlow(mdp.parallel.ParallelCheckpointFlow,
                                ParallelBiFlow, BiCheckpointFlow):
     """Parallel version of CheckpointFlow.
-    
+
     Can be used for saving intermediate results.
     """
-    
-    def train(self, data_iterables, checkpoints, msg_iterables=None, 
+
+    def train(self, data_iterables, checkpoints, msg_iterables=None,
               stop_messages=None,
               scheduler=None,
               train_callable_class=None,
               overwrite_result_container=True,
               **kwargs):
         """Train all trainable nodes in the flow.
-        
+
         Same as the train method in ParallelFlow, but with additional support
         of checkpoint functions as in CheckpointFlow.
         """
@@ -709,12 +614,12 @@ class ParallelCheckpointBiFlow(mdp.parallel.ParallelCheckpointFlow,
         kwargs["checkpoints"] = checkpoints
         super(ParallelCheckpointBiFlow, self).train(
                         data_iterables=data_iterables,
-                        scheduler=scheduler, 
+                        scheduler=scheduler,
                         train_callable_class=train_callable_class,
                         overwrite_result_container=overwrite_result_container,
                         msg_iterables=msg_iterables,
                         **kwargs)
-    
+
     def setup_parallel_training(self, data_iterables, checkpoints,
                                 msg_iterables=None,
                                 train_callable_class=BiFlowTrainCallable,
@@ -727,5 +632,3 @@ class ParallelCheckpointBiFlow(mdp.parallel.ParallelCheckpointFlow,
                                     train_callable_class=train_callable_class,
                                     msg_iterables=msg_iterables,
                                     **kwargs)
-        
-
